/*
   Copyright (c) 2014 by Plexim GmbH
   All rights reserved.

   A free license is granted to anyone to use this software for any legal
   non safety-critical purpose, including commercial applications, provided
   that:
   1) IT IS NOT USED TO DIRECTLY OR INDIRECTLY COMPETE WITH PLEXIM, and
   2) THIS COPYRIGHT NOTICE IS PRESERVED in its entirety.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
 */

#include "includes.h"

// code that needs to be loaded into RAM
extern Uint16 RamfuncsLoadStart, RamfuncsLoadEnd, RamfuncsRunStart;
// stack information (symbols generated by linker)
extern uint16_t stack_start, _STACK_SIZE;

// handles defined in main.cfg
extern const Semaphore_Handle BlinkSem;
extern const Timer_Handle TickTimer;
extern const Swi_Handle Swi0;

struct CTRL_DISPATCH_VARS CtrlDispatchVars;

// scope buffer
#define SCOPE_BUF_SIZE 1000
uint16_t ScopeBuffer[SCOPE_BUF_SIZE];

// stack monitor
struct STACK_MONITOR {
	uint32_t top;
	uint32_t bottom; // note: higher address than top
	uint16_t *ptr;
	uint32_t free;
};
struct STACK_MONITOR StackMon;

/**
 * Main interrupt routine
 */
Void ControlDispatcherHwi(UArg arg){
	PIL_beginInterruptCall();

	if(ProcessInputs()){
		// fast control task
		ControlTask1();
		// slow control task
		if(CtrlDispatchVars.task2DispatchTimer == 0)
		{
			Swi_post(Swi0); // this will call ControlTask2();
		}
	}

	CtrlDispatchVars.tickCounter++;
	//PIL_SCOPE_sample();

    // post the semaphore for LED blinking
	if(CtrlDispatchVars.ledBlinkTimer == 0){
		Semaphore_post(BlinkSem);
	}

	// update timers
	CtrlDispatchVars.task2DispatchTimer++;
	if(CtrlDispatchVars.task2DispatchTimer >= TASK2_PERIOD){
		CtrlDispatchVars.task2DispatchTimer = 0;
		CtrlDispatchVars.backgroundLoopCounter = 0;
	}

	CtrlDispatchVars.ledBlinkTimer++;
	if(CtrlDispatchVars.ledBlinkTimer >= LED_PERIOD){
		CtrlDispatchVars.ledBlinkTimer = 0;
	}

	// diagnostics
	uint32_t timer = Timer_getCount(TickTimer);
    if(timer > CtrlDispatchVars.tickTimerMax){
    	CtrlDispatchVars.tickTimerMax = timer;
    }
}

/**
 * Software interrupt for slow control task
 */
Void ControlDispatcherSwi0(UArg arg0, UArg arg1)
{
	CtrlDispatchVars.task2Active = TRUE;
	PLX_DELAY_US(300); // we add some fake "load" to demonstrate Task2 being interrupted by Task1
	ControlTask2();
	CtrlDispatchVars.task2Active = FALSE;
}

void ResetControlDispatcher()
{
	CtrlDispatchVars.tickCounter = 0;
	CtrlDispatchVars.task2DispatchTimer = 0;
	CtrlDispatchVars.ledBlinkTimer = 0;
}

void InitControlDispatcher()
{
	CtrlDispatchVars.task2Active = FALSE;
	CtrlDispatchVars.tickTimerMax = 0;

	ResetControlDispatcher();
}

/**
 * LED blinking using "task" and "semaphore" concepts
 */
Void BlinkTaskFxn(Void){
	static char hi = 0;
    while (TRUE) {
    	// pend on semaphore
        Semaphore_pend(BlinkSem, BIOS_WAIT_FOREVER);
        // blink LED
        if(hi){
        	hi = 0;
        	GpioG1DataRegs.GPCDAT.bit.GPIO70 = 0;
        } else {
        	hi = 1;
        	GpioG1DataRegs.GPCDAT.bit.GPIO70 = 1;
        }
    }
}

void InitStackMonitor()
{
	// init stack diagnostics
	StackMon.top = (uint32_t)(&stack_start);
	StackMon.bottom =  StackMon.top + (uint32_t)(&_STACK_SIZE)-1;
	StackMon.ptr = (uint16_t *)StackMon.bottom;
	StackMon.free = (uint32_t)(&_STACK_SIZE);
}

void RunStackMonitor()
{
	if(*StackMon.ptr == 0xBEBE)
	{
		// tag still present
		if((uint32_t)StackMon.ptr > StackMon.top)
		{
			StackMon.ptr--;
		}
		else
		{
			StackMon.ptr = (uint16_t*)StackMon.bottom; // reset search
		}
	}
	else
	{
		// this location has been used
		StackMon.free = StackMon.bottom - (uint32_t)StackMon.ptr;
		StackMon.ptr = (uint16_t*)StackMon.bottom; // reset search
	}
}

/**
 * Background task - handles communication with client
 */
Void BackgroundTaskFxn(Void){
	CtrlDispatchVars.backgroundLoopCounter++;
	PIL_backgroundCall();
	RunStackMonitor();
}

void MemCopy(Uint16 *SourceAddr, Uint16* SourceEndAddr, Uint16* DestAddr)
{
	while(SourceAddr < SourceEndAddr)
	{
	   *DestAddr++ = *SourceAddr++;
	}
	return;
}

Int main(int argc, char* argv[]){
	MemCopy(&RamfuncsLoadStart, &RamfuncsLoadEnd, &RamfuncsRunStart);

	InitControlDispatcher();
	InitStackMonitor();

	InitGPIO();
	InitCalibrations();

#ifdef USE_SCI_LINK
	 SCIInit(ONE_STOPBIT+ NO_PARITY + EIGHT_BITS, BAUD_RATE, LSPCLK_HZ);
	 SCIWriteString("\n\r\n\rStarting PIL Demo Software...");
	 SCIWriteString("\n\r");
#endif

    //initialize PIL
	PIL_init();
#if defined(USE_SCI_LINK)
	PIL_setLinkParams((unsigned char*)&PIL_D_Guid[0], (PIL_CommCallbackPtr_t)SCIPoll);
	PIL_setNodeAddress(PIL_D_StationAddress);
#elif defined(USE_DIRECT_LINK)
	PIL_setLinkParams((unsigned char*)&PIL_D_Guid[0], (PIL_CommCallbackPtr_t)IPCPoll);
	PIL_setBackgroundCommCallback((PIL_CommCallbackPtr_t)IPCPollBackground);
#else
#error Unsupported communication link!
#endif
	PIL_setCtrlCallback((PIL_CtrlCallbackPtr_t)PilCallback);
	//PIL_setScopeBuffer((uint16_t*)&ScopeBuffer, SCOPE_BUF_SIZE);

	PilInitOverrideProbes();
	InitControl();

	// get ready for PIL
	PIL_requestReadyMode();

    BIOS_start(); // will not return

    return (0);
}

